/* Copyright 2000 by the Massachusetts Institute of Technology.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both that copyright
 * notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in
 * advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 * M.I.T. makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */

#include "nmcalc.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

struct mask *mask_from_quadstring(char *in) {
  struct mask *out;
  char *ptr1, *ptr2;
  int a, b, c, d, addr, i;

  if (!in) return(NULL);
  
  ptr1=ptr2=in;
  ptr2=strchr(ptr1, '.');
  if (!ptr2) {
    printf("Missing octet for dotted quad notation for '%s'.\n", in);
    return(NULL);
  }
  a=atoi(ptr1);

  ptr1=ptr2+1;
  ptr2=strchr(ptr1, '.');
  if (!ptr2) {
    printf("Missing octet for dotted quad notation for '%s'.\n", in);
    return(NULL);
  }
  b=atoi(ptr1);

  ptr1=ptr2+1;
  ptr2=strchr(ptr1, '.');
  if (!ptr2) {
    printf("Missing octet for dotted quad notation for '%s'.\n", in);
    return(NULL);
  }
  c=atoi(ptr1);

  ptr1=ptr2+1;
  d=atoi(ptr1);

  if (a<0 || a>255 ||
      b<0 || b>255 ||
      c<0 || c>255 ||
      d<0 || d>255) {
    printf("Octet values must be bewteen 0 and 255.\n");
    return(NULL);
  }

  addr=d+(c*256)+(b*256*256)+(a*256*256*256);

  out=malloc(sizeof(struct mask));
  for (i=0; i<32; i++) {
    out->bits[i]=(addr & 1);
    addr=(addr >> 1);
  }
  
  return(out);
}

struct mask *mask_from_cidrstring(char *in) {
  struct mask *out;
  int i, cidr;
  
  if (in[0]=='/') in++;
  cidr=atoi(in);

  if (cidr<1 || cidr>32) {
    printf("CIDR block size must be between 1 and 32\n");
    return(NULL);
  }

  out=malloc(sizeof(struct mask));
  for (i=0; i<(32-cidr); i++) {
    out->bits[i]=0;
  }
  for (i=32-cidr; i<32; i++) {
    out->bits[i]=1;
  }
  return(out);
}

struct mask *mask_from_bitstring(char *in) {
  struct mask *out;
  int i;
  
  if (strlen(in)!=32) {
    printf("Need 32 bits in a bitstring\n");
    return(NULL);
  }

  out=malloc(sizeof(struct mask));
  for (i=31; i>=0; i--) {
    if (in[i] == '0') {
      out->bits[31-i]=0;
    } else if (in[i] == '1') {
      out->bits[31-i]=1;
    } else {
      printf("Only 1's and 0's may be in a bitstring\n");
      free(out);
      return(NULL);
    }
  }
  return(out);
}

char *mask_to_quadstring(struct mask *in) {
  int a, b, c, d;
  char *out;

  d=in->bits[0]+(2*in->bits[1])+(4*in->bits[2])+(8*in->bits[3])+(16*in->bits[4])+(32*in->bits[5])+(64*in->bits[6])+(128*in->bits[7]);
  c=in->bits[8]+(2*in->bits[9])+(4*in->bits[10])+(8*in->bits[11])+(16*in->bits[12])+(32*in->bits[13])+(64*in->bits[14])+(128*in->bits[15]);
  b=in->bits[16]+(2*in->bits[17])+(4*in->bits[18])+(8*in->bits[19])+(16*in->bits[20])+(32*in->bits[21])+(64*in->bits[22])+(128*in->bits[23]);
  a=in->bits[24]+(2*in->bits[25])+(4*in->bits[26])+(8*in->bits[27])+(16*in->bits[28])+(32*in->bits[29])+(64*in->bits[30])+(128*in->bits[31]);

  out=malloc(20);
  sprintf(out, "%i.%i.%i.%i", a, b, c, d);

  return(out);
}


char *mask_to_wildstring(struct mask *in) {
  int a, b, c, d;
  char *out;

  d=in->bits[0]+(2*in->bits[1])+(4*in->bits[2])+(8*in->bits[3])+(16*in->bits[4])+(32*in->bits[5])+(64*in->bits[6])+(128*in->bits[7]);
  c=in->bits[8]+(2*in->bits[9])+(4*in->bits[10])+(8*in->bits[11])+(16*in->bits[12])+(32*in->bits[13])+(64*in->bits[14])+(128*in->bits[15]);
  b=in->bits[16]+(2*in->bits[17])+(4*in->bits[18])+(8*in->bits[19])+(16*in->bits[20])+(32*in->bits[21])+(64*in->bits[22])+(128*in->bits[23]);
  a=in->bits[24]+(2*in->bits[25])+(4*in->bits[26])+(8*in->bits[27])+(16*in->bits[28])+(32*in->bits[29])+(64*in->bits[30])+(128*in->bits[31]);

  d=255-d;
  c=255-c;
  b=255-b;
  a=255-a;
  
  out=malloc(20);
  sprintf(out, "%i.%i.%i.%i", a, b, c, d);

  return(out);
}

char *mask_to_cidrstring(struct mask *in) {
  int i, j;
  char *out;

  i=j=0;
  
  while (in->bits[i]==0) {
    i++;
    if (i>31) break;
  }
  while (in->bits[i]==1) {
    i++;
    j++;
    if (i>31) break;
  }

  if (i!=32) {
    return(NULL);
  }

  out=malloc(20);
  sprintf(out, "/%i", j);

  return(out);
}

char *mask_to_bitstring(struct mask *in) {
  int i;
  char *out;

  out=malloc(35);
  strcpy(out, "");
  
  for (i=31; i>=0; i--) {
    if (in->bits[i]==0) {
      strcat(out, "0");
    } else {
      strcat(out, "1");
    }
  }
  return(out);
}

char *mask_to_nicebitstring(struct mask *in) {
  int i;
  char *out;

  out=malloc(35);
  strcpy(out, "");
  
  for (i=31; i>=0; i--) {
    if (in->bits[i]==0) {
      strcat(out, "0");
    } else {
      strcat(out, "1");
    }
    if (i%8==0) strcat(out, " ");
  }
  return(out);
}

char *mask_netlowaddr(struct mask *in, char *ip) {
  struct mask *m;
  char *ipstr, *maskstr, lowstr[35];
  int i;

  m=mask_from_quadstring(ip);
  if (!m) {
    return(NULL);
  }

  ipstr=mask_to_bitstring(m);
  maskstr=mask_to_bitstring(in);

  for (i=0; i<32; i++) {
    /* if it's part of the netmask then keep it */
    if (maskstr[i] == '1') {
      lowstr[i]=ipstr[i];
    } else {
      lowstr[i]='0';
    }
  }
  lowstr[32]='\0';
  return(mask_to_quadstring(mask_from_bitstring(lowstr)));
}

char *mask_nethighaddr(struct mask *in, char *ip) {
  struct mask *m;
  char *ipstr, *maskstr, lowstr[35];
  int i;

  m=mask_from_quadstring(ip);
  if (!m) {
    return(NULL);
  }
     
  ipstr=mask_to_bitstring(m);
  maskstr=mask_to_bitstring(in);

  for (i=0; i<32; i++) {
    /* if it's part of the netmask then keep it */
    if (maskstr[i] == '1') {
      lowstr[i]=ipstr[i];
    } else {
      lowstr[i]='1';
    }
  }
  lowstr[32]='\0';
    
  return(mask_to_quadstring(mask_from_bitstring(lowstr)));
}

int mask_num_addresses(struct mask *in) {
  int i, b;

  b=0;
  for (i=31; i>=0; i--) {
    if (in->bits[i]==0) b++;
  }
  return(pow(2,b));
}
